# This file is part of ranger, the console file manager.
# License: GNU GPL version 3, see the file "AUTHORS" for details.

"""Contains abbreviations to curses color/attribute constants.

Multiple attributes can be combined with the | (or) operator, toggled
with ^ (xor) and checked for with & (and). Examples:

attr = bold | underline
attr |= reverse
bool(attr & reverse) # => True
attr ^= reverse
bool(attr & reverse) # => False
"""

import curses

DEFAULT_FOREGROUND = curses.COLOR_WHITE
DEFAULT_BACKGROUND = curses.COLOR_BLACK
COLOR_PAIRS = {10: 0}


def get_color(fg, bg):
    """Returns the curses color pair for the given fg/bg combination."""

    key = (fg, bg)
    if key not in COLOR_PAIRS:
        size = len(COLOR_PAIRS)
        try:
            curses.init_pair(size, fg, bg)
        except curses.error:
            # If curses.use_default_colors() failed during the initialization
            # of curses, then using -1 as fg or bg will fail as well, which
            # we need to handle with fallback-defaults:
            if fg == -1:  # -1 is the "default" color
                fg = DEFAULT_FOREGROUND
            if bg == -1:  # -1 is the "default" color
                bg = DEFAULT_BACKGROUND

            try:
                curses.init_pair(size, fg, bg)
            except curses.error:
                # If this fails too, colors are probably not supported
                pass
        COLOR_PAIRS[key] = size

    return COLOR_PAIRS[key]


# pylint: disable=invalid-name,bad-whitespace
black = curses.COLOR_BLACK
blue = curses.COLOR_BLUE
cyan = curses.COLOR_CYAN
green = curses.COLOR_GREEN
magenta = curses.COLOR_MAGENTA
red = curses.COLOR_RED
white = curses.COLOR_WHITE
yellow = curses.COLOR_YELLOW
default = -1

normal = curses.A_NORMAL
bold = curses.A_BOLD
blink = curses.A_BLINK
reverse = curses.A_REVERSE
underline = curses.A_UNDERLINE
invisible = curses.A_INVIS
dim = curses.A_DIM

default_colors = (default, default, normal)
# pylint: enable=invalid-name,bad-whitespace

curses.setupterm()
# Adding BRIGHT to a color achieves what `bold` was used for.
BRIGHT = 8 if curses.tigetnum("colors") >= 16 else 0



# Copyright (c) 2018, Toby Slight. All rights reserved.
# ISC License (ISCL) - see LICENSE file for details.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cansi",
    version="0.0.3",
    author="Toby Slight",
    author_email="tslight@pm.me",
    description="Curses ANSI Parser",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tslight/cansi",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
    ),
)



import curses
import re

# pylint: disable=too-many-instance-attributes, too-few-public-methods, too-many-statements


class Cansi:
    """
    Curses Ansi Parser
    """

    def __init__(self, window):
        assert curses.has_colors(), "Curses wasn't configured to support colors."
        curses.use_default_colors()  # https://stackoverflow.com/a/44015131

        self.window = window

        for i in range(1, 8):
            curses.init_pair(i, i, -1)
            curses.init_pair(i + 7, curses.COLOR_WHITE, i)
            curses.init_pair(i + 14, curses.COLOR_BLACK, i)

        self.red_black = curses.color_pair(1)
        self.green_black = curses.color_pair(2)
        self.yellow_black = curses.color_pair(3)
        self.blue_black = curses.color_pair(4)
        self.magenta_black = curses.color_pair(5)
        self.cyan_black = curses.color_pair(6)
        self.white_black = curses.color_pair(7)
        self.white_red = curses.color_pair(8)
        self.white_green = curses.color_pair(9)
        self.white_yellow = curses.color_pair(10)
        self.white_blue = curses.color_pair(11)
        self.white_magenta = curses.color_pair(12)
        self.white_cyan = curses.color_pair(13)
        self.white_white = curses.color_pair(14)
        self.black_red = curses.color_pair(15)
        self.black_green = curses.color_pair(16)
        self.black_yellow = curses.color_pair(17)
        self.black_blue = curses.color_pair(18)
        self.black_magenta = curses.color_pair(19)
        self.black_cyan = curses.color_pair(20)
        self.black_white = curses.color_pair(21)

        self.red_black_bold = curses.color_pair(1) | curses.A_BOLD
        self.green_black_bold = curses.color_pair(2) | curses.A_BOLD
        self.yellow_black_bold = curses.color_pair(3) | curses.A_BOLD
        self.blue_black_bold = curses.color_pair(4) | curses.A_BOLD
        self.magenta_black_bold = curses.color_pair(5) | curses.A_BOLD
        self.cyan_black_bold = curses.color_pair(6) | curses.A_BOLD
        self.white_black_bold = curses.color_pair(7) | curses.A_BOLD
        self.white_red_bold = curses.color_pair(8) | curses.A_BOLD
        self.white_green_bold = curses.color_pair(9) | curses.A_BOLD
        self.white_yellow_bold = curses.color_pair(10) | curses.A_BOLD
        self.white_blue_bold = curses.color_pair(11) | curses.A_BOLD
        self.white_magenta_bold = curses.color_pair(12) | curses.A_BOLD
        self.white_cyan_bold = curses.color_pair(13) | curses.A_BOLD
        self.white_white_bold = curses.color_pair(14) | curses.A_BOLD
        self.black_red_bold = curses.color_pair(15) | curses.A_BOLD
        self.black_green_bold = curses.color_pair(16) | curses.A_BOLD
        self.black_yellow_bold = curses.color_pair(17) | curses.A_BOLD
        self.black_blue_bold = curses.color_pair(18) | curses.A_BOLD
        self.black_magenta_bold = curses.color_pair(19) | curses.A_BOLD
        self.black_cyan_bold = curses.color_pair(20) | curses.A_BOLD
        self.black_white_bold = curses.color_pair(21) | curses.A_BOLD

        self.ansi_attr = {
            "1": curses.A_BOLD,
            "4": curses.A_UNDERLINE,
            "5": curses.A_BLINK,
            "7": curses.A_REVERSE,
        }

        self.ansi_color = {
            "0": self.white_black,
            "0;": self.white_black,
            "30": self.white_black,
            "31": self.red_black,
            "32": self.green_black,
            "33": self.yellow_black,
            "34": self.blue_black,
            "35": self.magenta_black,
            "36": self.cyan_black,
            "37": self.white_black,
            "90": self.white_black_bold,
            "91": self.red_black_bold,
            "92": self.green_black_bold,
            "93": self.yellow_black_bold,
            "94": self.blue_black_bold,
            "95": self.magenta_black_bold,
            "96": self.cyan_black_bold,
            "97": self.white_black_bold,
            "0;30": self.white_black,
            "0;31": self.red_black,
            "0;32": self.green_black,
            "0;33": self.yellow_black,
            "0;34": self.blue_black,
            "0;35": self.magenta_black,
            "0;36": self.cyan_black,
            "0;37": self.white_black,
            "0;90": self.white_black_bold,
            "0;91": self.red_black_bold,
            "0;92": self.green_black_bold,
            "0;93": self.yellow_black_bold,
            "0;94": self.blue_black_bold,
            "0;95": self.magenta_black_bold,
            "0;96": self.cyan_black_bold,
            "0;97": self.white_black_bold,
            "1;30": self.white_black_bold,
            "1;31": self.red_black_bold,
            "1;32": self.green_black_bold,
            "1;33": self.yellow_black_bold,
            "1;34": self.blue_black_bold,
            "1;35": self.magenta_black_bold,
            "1;36": self.cyan_black_bold,
            "1;37": self.white_black_bold,
            "30;0": self.white_black,
            "31;0": self.red_black,
            "32;0": self.green_black,
            "33;0": self.yellow_black,
            "34;0": self.blue_black,
            "35;0": self.magenta_black,
            "36;0": self.cyan_black,
            "37;0": self.white_black,
            "30;1": self.white_black_bold,
            "31;1": self.red_black_bold,
            "32;1": self.green_black_bold,
            "33;1": self.yellow_black_bold,
            "34;1": self.blue_black_bold,
            "35;1": self.magenta_black_bold,
            "36;1": self.cyan_black_bold,
            "37;1": self.white_black_bold,
        }

    def addstr(self, y, x, string):
        """
        Adds the color-formatted string to the given window, in the given
        coordinates

        ANSI escapes (CSI - Control Sequence Introducers) always start with
        \\e, \033 (octal), \x1b (hex) or \001b. They’re just various ways of
        inserting byte 27 into a string. In an ASCII table, 0x1b is called ESC,
        and this is why.
        """
        ansi_split = string.split("\x1b[")
        color_pair = self.white_black

        # Print the first part of the string without color change
        self.window.addstr(y, x, ansi_split[0], color_pair)
        x += len(ansi_split[0])

        # Iterate over the rest of the string-parts and print them with their colors
        for substring in ansi_split[1:]:
            if substring.startswith("0K"):
                substring = substring.replace("0K", "")
            elif substring.startswith("1G"):
                x = 0
            else:
                ansi_code = substring.split("m")[0]
                substring = substring[len(ansi_code) + 1 :]
                if ansi_code in ["1", "4", "5", "7", "8"]:
                    color_pair = color_pair | self.ansi_attr[ansi_code]
                else:
                    color_pair = self.ansi_color[ansi_code]

            if substring:
                self.window.addstr(y, x, substring, color_pair)
                x += len(substring)



# This file is part of ranger, the console file manager.
# License: GNU GPL version 3, see the file "AUTHORS" for details.
# Author: David Barnett <davidbarnett2@gmail.com>, 2010

"""A library to help to convert ANSI codes to curses instructions."""

import re

from color import color


# pylint: disable=invalid-name
ansi_re = re.compile("(\x1b" + r"\[\d*(?:;\d+)*?[a-zA-Z])")
codesplit_re = re.compile(r"38;5;(\d+);|48;5;(\d+);|(\d*);")
reset = "\x1b[0m"
# pylint: enable=invalid-name


def split_ansi_from_text(ansi_text):
    return ansi_re.split(ansi_text)


# For information on the ANSI codes see
# githttp://en.wikipedia.org/wiki/ANSI_escape_code


def text_with_fg_bg_attr(
    ansi_text,
):  # pylint: disable=too-many-branches,too-many-statements
    fg, bg, attr = -1, -1, 0
    for chunk in split_ansi_from_text(ansi_text):
        if chunk and chunk[0] == "\x1b":
            if chunk[-1] != "m":
                continue
            match = re.match(r"^.\[(.*).$", chunk)
            if not match:
                # XXX I have no test case to determine what should happen here
                continue
            attr_args = match.group(1)

            # Convert arguments to attributes/colors
            for x256fg, x256bg, arg in codesplit_re.findall(attr_args + ";"):
                # first handle xterm256 codes
                try:
                    if x256fg:  # xterm256 foreground
                        fg = int(x256fg)
                        continue
                    elif x256bg:  # xterm256 background
                        bg = int(x256bg)
                        continue
                    elif arg:  # usual ansi code
                        n = int(arg)
                    else:  # empty code means reset
                        n = 0
                except ValueError:
                    continue

                if n == 0:  # reset colors and attributes
                    fg, bg, attr = -1, -1, 0

                elif n == 1:  # enable attribute
                    attr |= color.bold
                elif n == 4:
                    attr |= color.underline
                elif n == 5:
                    attr |= color.blink
                elif n == 7:
                    attr |= color.reverse
                elif n == 8:
                    attr |= color.invisible

                elif n == 22:  # disable attribute
                    attr &= not color.bold
                elif n == 24:
                    attr &= not color.underline
                elif n == 25:
                    attr &= not color.blink
                elif n == 27:
                    attr &= not color.reverse
                elif n == 28:
                    attr &= not color.invisible

                elif n >= 30 and n <= 37:  # 8 ansi foreground and background colors
                    fg = n - 30
                elif n == 39:
                    fg = -1
                elif n >= 40 and n <= 47:
                    bg = n - 40
                elif n == 49:
                    bg = -1

                # 8 aixterm high intensity colors (light but not bold)
                elif n >= 90 and n <= 97:
                    fg = n - 90 + 8
                elif n == 99:
                    fg = -1
                elif n >= 100 and n <= 107:
                    bg = n - 100 + 8
                elif n == 109:
                    bg = -1

            yield (fg, bg, attr)

        else:
            yield chunk


def char_len(ansi_text):
    """Count the number of visible characters.

    >>> char_len("\x1b[0;30;40mX\x1b[0m")
    1
    >>> char_len("\x1b[0;30;40mXY\x1b[0m")
    2
    >>> char_len("\x1b[0;30;40mX\x1b[0mY")
    2
    >>> char_len("hello")
    5
    >>> char_len("")
    0
    """
    return len((ansi_re.sub("", ansi_text)))


def char_slice(ansi_text, start, length):
    """Slices a string with respect to ansi code sequences

    Acts as if the ansi codes aren't there, slices the text from the
    given start point to the given length and adds the codes back in.

    >>> test_string = "abcde\x1b[30mfoo\x1b[31mbar\x1b[0mnormal"
    >>> split_ansi_from_text(test_string)
    ['abcde', '\\x1b[30m', 'foo', '\\x1b[31m', 'bar', '\\x1b[0m', 'normal']
    >>> char_slice(test_string, 1, 3)
    'bcd'
    >>> char_slice(test_string, 5, 6)
    '\\x1b[30mfoo\\x1b[31mbar'
    >>> char_slice(test_string, 0, 8)
    'abcde\\x1b[30mfoo'
    >>> char_slice(test_string, 4, 4)
    'e\\x1b[30mfoo'
    >>> char_slice(test_string, 11, 100)
    '\\x1b[0mnormal'
    >>> char_slice(test_string, 9, 100)
    '\\x1b[31mar\\x1b[0mnormal'
    >>> char_slice(test_string, 9, 4)
    '\\x1b[31mar\\x1b[0mno'
    """
    chunks = []
    last_color = ""
    pos = old_pos = 0
    for i, chunk in enumerate(split_ansi_from_text(ansi_text)):
        if i % 2 == 1:
            last_color = chunk
            continue

        old_pos = pos
        pos += len(chunk)
        if pos <= start:
            pass  # seek
        elif old_pos < start and pos >= start:
            chunks.append(last_color)
            chunks.append(str(chunk[start - old_pos : start - old_pos + length]))
        elif pos > length + start:
            chunks.append(last_color)
            chunks.append(str(chunk[: start - old_pos + length]))
        else:
            chunks.append(last_color)
            chunks.append(str(chunk))
        if pos - start >= length:
            break
    return "".join(chunks)
